@@ -1,49 +1,69 @@
-;;; hui.el --- GNU Emacs User Interface to Hyperbole
-
-;; Copyright (C) 1991-1995, 2008 Free Software Foundation, Inc.
-;; Developed with support from Motorola Inc.
-
-;; Author: Bob Weiner, Brown U.
-;; Maintainer: Mats Lidell <matsl@contactor.se>
-;; Keywords: hypermedia
-
+;;; hui.el ---  GNU Hyperbole button and hyperlink user interface
+;;
+;; Author:       Bob Weiner
+;;
+;; Orig-Date:    19-Sep-91 at 21:42:03
+;;
+;; Copyright (C) 1991-2016  Free Software Foundation, Inc.
+;; See the "HY-COPY" file for license information.
+;;
 ;; This file is part of GNU Hyperbole.
 
-;; GNU Hyperbole is free software; you can redistribute it and/or
-;; modify it under the terms of the GNU General Public License as
-;; published by the Free Software Foundation; either version 3, or (at
-;; your option) any later version.
-
-;; GNU Hyperbole is distributed in the hope that it will be useful,
-;; but WITHOUT ANY WARRANTY; without even the implied warranty of
-;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-;; General Public License for more details.
-
-;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-;; Boston, MA 02110-1301, USA.
-
 ;;; Commentary:
 
 ;;; Code:
-
-;;;
+;;; ************************************************************************
 ;;; Other required Elisp libraries
-;;;
+;;; ************************************************************************
 
-(require 'hargs) (require 'set) (require 'hmail)
+(require 'hargs)
+(require 'set)
+(require 'hmail)
 
-;;;
+;;; ************************************************************************
 ;;; Public variables
-;;;
+;;; ************************************************************************
 
-(defvar hui:ebut-delete-confirm-p t
-  "*Non-nil means prompt before interactively deleting explicit buttons.")
+(defcustom hui:ebut-delete-confirm-p t
+  "*Non-nil means prompt before interactively deleting explicit buttons."
+  :type 'boolean
+  :group 'hyperbole-buttons)
 
-;;;
+(defcustom hui:ebut-prompt-for-action nil
+  "*Non-nil means prompt for a button-specific action when creating buttons."
+  :type 'boolean
+  :group 'hyperbole-buttons)
+
+;;; ************************************************************************
 ;;; Public functions
-;;;
+;;; ************************************************************************
+
+(defun hui:bind-key (cmd &optional new-key)
+  "Remove existing global key binding for CMD, rebind it to optional NEW-KEY (prompted for) and confirm the change."
+  (interactive "CCommand to change key binding of: \nKNew key to bind: ")
+  (if (not (functionp cmd))
+      (error "(hui:bind-key): Invalid command, `%s'" cmd))
+  (let* ((old-key (where-is-internal cmd global-map t))
+	 ;; Force multi-character key sequences to echo in the minibuffer
+	 (echo-keystrokes 1)
+	 old-key-text
+	 new-key-text)
+    (if old-key (setq old-key-text (key-description old-key)))
+    (if (null new-key)
+	(setq new-key
+	      (with-selected-window (minibuffer-window)
+		(read-key-sequence
+		 (if old-key
+		     (format "{%s} runs `%s'; change it to key: " old-key-text cmd)
+		   (format "New key to run `%s': " cmd))))))
+    (cond ((equal new-key (kbd "\C-g"))
+	   (keyboard-quit))
+	  (new-key (global-set-key new-key cmd)
+		   (setq new-key-text (key-description new-key))))
+    (if old-key
+	(progn (global-unset-key old-key)
+	       (message "{%s} now runs `%s'; prior {%s} binding removed" new-key-text cmd old-key-text))
+      (message "{%s} now runs `%s'" new-key-text cmd))))
 
 (defun hui:ebut-create (&optional start end)
   "Creates an explicit but starting from label between optional START and END.
@@ -56,7 +76,7 @@ label."
   (let ((default-lbl) lbl but-buf actype)
     (save-excursion
       (setq default-lbl
-	    (hui:hbut-label-default start end (not (interactive-p)))
+	    (hui:hbut-label-default start end (not (called-interactively-p 'interactive)))
 	    lbl (hui:hbut-label default-lbl "ebut-create"))
       (if (not (equal lbl default-lbl)) (setq default-lbl nil))
       
@@ -69,8 +89,7 @@ label."
       (hattr:set 'hbut:current 'actype actype)
       (hattr:set 'hbut:current 'args (hargs:actype-get actype))
       (hattr:set 'hbut:current 'action
-		 (and (boundp 'hui:ebut-prompt-for-action)
-		      hui:ebut-prompt-for-action (hui:action actype)))
+		 (and hui:ebut-prompt-for-action (hui:action actype)))
       )
     (ebut:operate lbl nil)))
 
@@ -90,7 +109,7 @@ Signals an error if point is not within a button."
 	((not (stringp but-key))
 	 (hypb:error
 	  "(ebut-delete): Invalid label key argument: '%s'." but-key)))
-  (let ((interactive (interactive-p)))
+  (let ((interactive (called-interactively-p 'interactive)))
     (if (and hui:ebut-delete-confirm-p interactive)
 	(if (y-or-n-p (format "Delete button %s%s%s? "
 			      ebut:start
@@ -133,7 +152,7 @@ Signals an error when no such button is found in the current buffer."
 	(but-buf (current-buffer))
 	actype but new-lbl)
     (save-excursion
-      (or (interactive-p)
+      (or (called-interactively-p 'interactive)
 	  (hui:buf-writable-err but-buf "ebut-modify"))
       
       (or (setq but (ebut:get lbl-key but-buf))
@@ -143,9 +162,8 @@ Signals an error when no such button is found in the current buffer."
       (setq new-lbl
 	    (hargs:read
 	     "Change button label to: "
-	     (function
-	       (lambda (lbl)
-		(and (not (string= lbl "")) (<= (length lbl) ebut:max-len))))
+	     (lambda (lbl)
+	       (and (not (string= lbl "")) (<= (length lbl) ebut:max-len)))
 	     lbl
 	     (format "(ebut-modify): Enter a string of at most %s chars."
 		     ebut:max-len)
@@ -157,8 +175,7 @@ Signals an error when no such button is found in the current buffer."
       (hattr:set 'hbut:current 'actype actype)
       (hattr:set 'hbut:current 'args (hargs:actype-get actype 'modifying))
       (hattr:set 'hbut:current 'action
-		 (and (boundp 'hui:ebut-prompt-for-action)
-		      hui:ebut-prompt-for-action (hui:action actype)))
+		 (and hui:ebut-prompt-for-action (hui:action actype)))
       )
     (ebut:operate lbl new-lbl)))
 
@@ -190,10 +207,9 @@ Signals an error if any problem occurs."
 			 (setq new-label
 			       (hargs:read
 				"Rename button label to: "
-				(function
-				  (lambda (lbl)
-				   (and (not (string= lbl ""))
-					(<= (length lbl) ebut:max-len))))
+			        (lambda (lbl)
+				  (and (not (string= lbl ""))
+				       (<= (length lbl) ebut:max-len)))
 				curr-label
 				(format
 				 "(ebut-rename): Use a quoted string of at most %s chars."
@@ -202,7 +218,7 @@ Signals an error if any problem occurs."
        (list curr-label new-label))))
 
   (save-excursion
-    (if (interactive-p)
+    (if (called-interactively-p 'interactive)
 	nil
       (hui:buf-writable-err (current-buffer) "ebut-rename")
       (if (or (not (stringp curr-label)) (string= curr-label ""))
@@ -252,10 +268,10 @@ a menu to find any of the occurrences."
 					(concat ebut:start string ebut:end)))))
 	  (goto-char (point-min))
 	  (pop-to-buffer out-buf)
-	  (if (interactive-p) (message "%d match%s." total
+	  (if (called-interactively-p 'interactive) (message "%d match%s." total
 				       (if (> total 1) "es" ""))
 	    total))
-      (if (interactive-p) (message "No matches.")
+      (if (called-interactively-p 'interactive) (message "No matches.")
 	total))))
 
 (defun hui:error (&rest args)
@@ -277,8 +293,7 @@ a menu to find any of the occurrences."
 	(hattr:set 'hbut:current 'actype actype)
 	(hattr:set 'hbut:current 'args (hargs:actype-get actype))
 	(hattr:set 'hbut:current 'action
-		   (and (boundp 'hui:ebut-prompt-for-action)
-			hui:ebut-prompt-for-action (hui:action actype)))
+		   (and hui:ebut-prompt-for-action (hui:action actype)))
 	(setq lbl (concat lbl (ebut:operate lbl nil)))
 	(goto-char (point-max))
 	(insert "\n")
@@ -295,13 +310,13 @@ Signals an error when no such button is found."
 			(find-file-noselect gbut:file) "gbut-modify")
 		       (hbut:label-to-key
 			(hargs:read-match "Global button to modify: "
-					  (mapcar 'list (gbut:lbl-list))
+					  (mapcar 'list (gbut:label-list))
 					  nil t nil 'ebut)))))
   (let ((lbl (hbut:key-to-label lbl-key))
 	(but-buf (find-file-noselect gbut:file))
 	actype but new-lbl)
     (save-excursion
-      (or (interactive-p)
+      (or (called-interactively-p 'interactive)
 	  (hui:buf-writable-err but-buf "gbut-modify"))
       
       (or (setq but (ebut:get lbl-key but-buf))
@@ -312,9 +327,8 @@ Signals an error when no such button is found."
       (setq new-lbl
 	    (hargs:read
 	     "Change global button label to: "
-	     (function
-	       (lambda (lbl)
-		(and (not (string= lbl "")) (<= (length lbl) ebut:max-len))))
+	     (lambda (lbl)
+	       (and (not (string= lbl "")) (<= (length lbl) ebut:max-len)))
 	     lbl
 	     (format "(gbut-modify): Enter a string of at most %s chars."
 		     ebut:max-len)
@@ -326,8 +340,7 @@ Signals an error when no such button is found."
       (hattr:set 'hbut:current 'actype actype)
       (hattr:set 'hbut:current 'args (hargs:actype-get actype 'modifying))
       (hattr:set 'hbut:current 'action
-		 (and (boundp 'hui:ebut-prompt-for-action)
-		      hui:ebut-prompt-for-action (hui:action actype)))
+		 (and hui:ebut-prompt-for-action (hui:action actype)))
       (set-buffer but-buf)
       (ebut:operate lbl new-lbl))))
 
@@ -343,12 +356,22 @@ Default is the current button."
 			(hargs:read-match "Button to execute: " lst nil t
 					  (ebut:label-p 'as-label) 'ebut))))
 	    (t (hypb:error "(hbut-act): No explicit buttons in buffer."))))))
-  (cond ((and (interactive-p) (null but))
+  (cond ((and (called-interactively-p 'interactive) (null but))
 	 (hypb:error "(hbut-act): No current button to activate."))
 	((not (hbut:is-p but))
 	 (hypb:error "(hbut-act): Button is invalid; it has no attributes."))
 	(t (or but (setq but 'hbut:current))
-	   (hui:but-flash) (hyperb:act but))))
+	   (hui:but-flash) (hbut:act but))))
+
+(defun hui:hbut-current-act ()
+  "Activate Hyperbole button at point or signal an error if there is no such button."
+  (interactive)
+  (let ((but (hbut:at-p)))
+    (cond ((null but)
+	   (hypb:error "(hbut-act): No current button to activate."))
+	  ((not (hbut:is-p but))
+	   (hypb:error "(hbut-act): Button is invalid; it has no attributes."))
+	  (t (hui:but-flash) (hbut:act but)))))
 
 (defun hui:hbut-help (&optional but)
   "Checks for and explains an optional button given by symbol, BUT.
@@ -364,17 +387,19 @@ BUT defaults to the button whose label point is within."
       (cond (but (hypb:error "(hbut-help): Invalid button."))
 	    (t   (hypb:error
 		  "(hbut-help): Not on an implicit button and no buffer explicit buttons."))))
-  (let* ((sym-name (htype:names 'ibtypes (hattr:get but 'categ)))
-	 (type-help-func (if sym-name (intern-soft (concat sym-name ":help")))))
+  (let ((type-help-func (intern-soft
+			 (concat 
+			  (htype:names 'ibtypes (hattr:get but 'categ))
+			  ":help"))))
     (or (equal (hypb:indirect-function 'hui:but-flash)
-	       (function (lambda nil)))
+	       (lambda nil))
 	;; Only flash button if point is on it.
 	(let ((lbl-key (hattr:get but 'lbl-key)))
 	  (and lbl-key
 	       (or (equal lbl-key (ebut:label-p))
 		   (equal lbl-key (ibut:label-p)))
 	       (hui:but-flash))))
-    (if type-help-func
+    (if (functionp type-help-func)
 	(funcall type-help-func but)
       (let ((total (hbut:report but)))
 	(if total (hui:help-ebut-highlight))))))
@@ -382,9 +407,8 @@ BUT defaults to the button whose label point is within."
 (defun hui:hbut-label (default-label func-name)
   "Reads button label from user using DEFAULT-LABEL and caller's FUNC-NAME."
   (hargs:read "Button label: "
-	      (function
-		(lambda (lbl)
-		  (and (not (string= lbl "")) (<= (length lbl) ebut:max-len))))
+	      (lambda (lbl)
+		(and (not (string= lbl "")) (<= (length lbl) ebut:max-len)))
 	      default-label
 	      (format "(%s): Enter a string of at most %s chars."
 		      func-name ebut:max-len)
@@ -392,7 +416,7 @@ BUT defaults to the button whose label point is within."
 
 (defun hui:hbut-label-default (start end &optional skip-len-test)
   "Returns default label based on START and END region markers or points.
-Optional SKIP-LEN-TEST means don't limit label to 'ebut:max-len' length.
+Optional SKIP-LEN-TEST means don't limit label to `ebut:max-len' length.
 Returns nil if START or END are invalid or if region fails length test. 
 
 Also has side effect of moving point to start of default label, if any."
@@ -416,11 +440,11 @@ See 'hbut:report'."
 	  (progn (hui:help-ebut-highlight)
 		 (message "%d button%s." total (if (/= total 1) "s" "")))))))
 
-(fset 'hui:hbut-summarize 'hui:hbut-report)
+(defalias 'hui:hbut-summarize 'hui:hbut-report)
 
 (defun hui:link-directly ()
   "Creates a Hyperbole link button at depress point, linked to release point.
-See also documentation for 'hui:link-possible-types'."
+See also documentation for `hui:link-possible-types'."
   (let* ((link-types (hui:link-possible-types))
 	 (but-window action-key-depress-window)
 	 (num-types (length link-types))
@@ -430,64 +454,64 @@ See also documentation for 'hui:link-possible-types'."
     (select-window action-key-depress-window)
     (hui:buf-writable-err (current-buffer) "link-directly")
     (if (ebut:at-p)
-	(progn
-	  (setq but-modify t
-		but-loc (hattr:get 'hbut:current 'loc)
-		but-dir (hattr:get 'hbut:current 'dir)
-		lbl-key (hattr:get 'hbut:current 'lbl-key)))
+	(setq but-modify t
+	      but-loc (hattr:get 'hbut:current 'loc)
+	      but-dir (hattr:get 'hbut:current 'dir)
+	      lbl-key (hattr:get 'hbut:current 'lbl-key))
       (setq but-loc (hui:key-src (current-buffer))
 	    but-dir (hui:key-dir (current-buffer))
 	    lbl-key (hbut:label-to-key
-		      (hui:hbut-label
-			(if (marker-position (hypb:mark-marker t))
-			    (hui:hbut-label-default
-			      (region-beginning) (region-end)))
-			"link-directly"))))
-
-    ;; XEmacs 21.5 thinks windows isn't live. Bug!?
-    (if (window-live-p release-window)
-	(select-window release-window))
+		     (hui:hbut-label
+		      (cond ((hmouse-prior-active-region)
+			     hkey-region)
+			    ((marker-position (hypb:mark-marker t))
+			     (hui:hbut-label-default
+			      (region-beginning) (region-end))))
+		      "link-directly"))))
+    (select-window release-window)
 
     (cond ((= num-types 0)
 	   (error "(link-directly): No possible link type to create."))
 	  ((= num-types 1)
-	   (hui:link-create but-modify
-			    but-window lbl-key but-loc but-dir
-			    (setq type-and-args (car link-types))))
-	  (t;; more than 1
+	   (setq type-and-args (hui:list-remove-text-properties (car link-types)))
+	   (hui:link-create but-modify but-window lbl-key but-loc but-dir type-and-args))
+	  (t ;; more than 1
 	    (let ((item)
 		  type)
+	      (setq type-and-args
+		    (hui:menu-select
+		     (cons '("Link to>")
+			   (mapcar
+			    (lambda (type-and-args)
+			      (setq type (car type-and-args))
+			      (list 
+			       (capitalize
+				(if (string-match
+				     "^\\(link-to\\|eval\\)-"
+				     (setq item (symbol-name type)))
+				    (setq item (substring
+						item (match-end 0)))
+				  item))
+			       type-and-args
+			       (documentation
+				(intern (concat "actypes::"
+						(symbol-name type))))))
+			    link-types)))
+		    type-and-args (hui:list-remove-text-properties type-and-args))
 	      (hui:link-create
 		but-modify but-window
-		lbl-key but-loc but-dir
-		(setq type-and-args
-		      (hui:menu-select
-			(cons '("Link to>")
-			      (mapcar
-				(function
-				  (lambda (type-and-args)
-				    (setq type (car type-and-args))
-				    (list 
-				      (capitalize
-					(if (string-match
-					      "^\\(link-to\\|eval\\)-"
-					      (setq item (symbol-name type)))
-					    (setq item (substring
-							 item (match-end 0)))
-					  item))
-				      type-and-args
-				      (documentation
-				       (intern (concat "actypes::"
-						       (symbol-name type)))))))
-				link-types))))))))
-    (message "`%s' button %s type `%s'."
+		lbl-key but-loc but-dir type-and-args))))
+    (message "`%s' button %s %s with %S."
 	     (hbut:key-to-label lbl-key)
-	     (if but-modify "set to" "created with")
-	     (car type-and-args))))
+	     (if but-modify "now executes" "executes")
+	     (car type-and-args)
+	     (if (= 1 (length (cdr type-and-args)))
+		 (cadr type-and-args)
+	       (cdr type-and-args)))))
 
-;;;
-;;; Private functions
-;;;
+;;; ************************************************************************
+;;; Private functions - used only within Hyperbole
+;;; ************************************************************************
 
 (defun hui:action (actype &optional prompt)
   "Prompts for and returns an action to override action from ACTYPE."
@@ -516,15 +540,14 @@ See also documentation for 'hui:link-possible-types'."
 		      ;; the user and provide another chance for entry.
 		      (not (memq t
 				 (mapcar
-				  (function
-				    (lambda (param)
-				     (setq param (symbol-name param))
-				     (and (string-match
-					   (concat "[\( \t\n,']"
-						   (regexp-quote param)
-						   "[\(\) \t\n\"]")
-					   act-str)
-					  t)))
+				  (lambda (param)
+				    (setq param (symbol-name param))
+				    (and (string-match
+					  (concat "[\( \t\n\r,']"
+						  (regexp-quote param)
+						  "[\(\) \t\n\r\"]")
+					  act-str)
+					 t))
 				  params)))
 		      ))
 	   (beep) (message "Action must use at least one parameter.")
@@ -532,9 +555,7 @@ See also documentation for 'hui:link-possible-types'."
 	 (let (head)
 	   (while (cond ((listp act)
 			 (and act (setq head (car act))
-			      (not (or (eq head 'lambda)
-				       (eq head 'defun)
-				       (eq head 'defmacro)))
+			      (not (memq head '(lambda defun defmacro defsubst defin)))
 			      (setq act (list 'lambda params act))
 			      nil  ;; terminate loop
 			      ))
@@ -565,7 +586,7 @@ DEFAULT-ACTYPE may be a valid symbol or symbol-name."
     ))
 
 (defun hui:buf-writable-err (but-buf func-name)
-  "If BUT-BUF is read-only or is unwritable, signal an error from FUNC-NAME."
+  "If BUT-BUF is read-only, signal an error from FUNC-NAME."
   (let ((obuf (prog1 (current-buffer) (set-buffer but-buf)))
 	;; (unwritable (and buffer-file-name
 	;;		 (not (file-writable-p buffer-file-name))))
@@ -578,13 +599,8 @@ DEFAULT-ACTYPE may be a valid symbol or symbol-name."
     ;;		     (file-name-nondirectory buffer-file-name))))
     (if buffer-read-only
 	(setq err
-	      (format
-	       "Button buffer '%s' is read-only.  Use %s to change it."
-	       (buffer-name but-buf)
-	       (hypb:cmd-key-string
-		(if (where-is-internal 'vc-toggle-read-only)
-		    'vc-toggle-read-only 'toggle-read-only))
-	       )))
+	      (format "Button buffer '%s' is read-only.  Use {%s} to change it."
+		      (buffer-name but-buf) (hmouse-read-only-toggle-key))))
     (set-buffer obuf)
     (if err (progn (pop-to-buffer but-buf) (hypb:error err)))))
 
@@ -598,14 +614,13 @@ DEFAULT-ACTYPE may be a valid symbol or symbol-name."
 		 (or prompt "Button's buffer: ")
 		 (delq nil
 		       (mapcar
-			(function
-			  (lambda (buf)
-			   (let ((b (buffer-name buf)))
-			     (if (and (not (string-match "mail\\*" b))
-				      (not (string-match "\\*post-news\\*" b))
-				      (string-match "\\`[* ]" b))
-				 nil 
-			       (cons b nil)))))
+		        (lambda (buf)
+			  (let ((b (buffer-name buf)))
+			    (if (and (not (string-match "mail\\*" b))
+				     (not (string-match "\\*post-news\\*" b))
+				     (string-match "\\`[* ]" b))
+				nil 
+			      (cons b nil))))
 			(buffer-list)))
 		 nil t (buffer-name) 'buffer))
 	  (or (null buf-name) (equal buf-name "")))
@@ -652,33 +667,32 @@ within."
   "Remove delimiters from button given by BUT-KEY in KEY-SRC of DIRECTORY.
 All args are optional, the current button and buffer file are the defaults."
   (interactive)
-  (let ((form (function
-	       (lambda ()
-		 (let ((buffer-read-only) start end)
-		   (setq start (match-beginning 0)
-			 end (match-end 0))
-		   (and (fboundp 'hproperty:but-delete)
-			(hproperty:but-delete start))
-		   (skip-chars-backward " \t\n")
-		   (skip-chars-backward "0-9")
-		   (if (= (preceding-char) (string-to-char ebut:instance-sep))
-		       (setq start (1- (point))))
-		   (if (search-backward ebut:start (- (point) ebut:max-len) t)
-		       (if current-prefix-arg
-			   ;; Remove button label, delimiters and preceding
-			   ;; space, if any.
-			   (delete-region (max (point-min)
-					       (1- (match-beginning 0)))
-					  end)
-			 ;;
-			 ;; Remove button delimiters only.
-			 ;;
-			 ;; Remove button ending delimiter
-			 (delete-region start end)
-			 ;; Remove button starting delimiter
-			 (delete-region (match-beginning 0)
-					(match-end 0)))))))))
-    (if (interactive-p)
+  (let ((form (lambda ()
+		(let ((buffer-read-only) start end)
+		  (setq start (match-beginning 0)
+			end (match-end 0))
+		  (and (fboundp 'hproperty:but-delete)
+		       (hproperty:but-delete start))
+		  (skip-chars-backward " \t\n\r")
+		  (skip-chars-backward "0-9")
+		  (if (eq (preceding-char) (string-to-char ebut:instance-sep))
+		      (setq start (1- (point))))
+		  (if (search-backward ebut:start (- (point) ebut:max-len) t)
+		      (if current-prefix-arg
+			  ;; Remove button label, delimiters and preceding
+			  ;; space, if any.
+			  (delete-region (max (point-min)
+					      (1- (match-beginning 0)))
+					 end)
+			;;
+			;; Remove button delimiters only.
+			;;
+			;; Remove button ending delimiter
+			(delete-region start end)
+			;; Remove button starting delimiter
+			(delete-region (match-beginning 0)
+				       (match-end 0))))))))
+    (if (called-interactively-p 'interactive)
 	(save-excursion
 	  (if (search-forward ebut:end nil t) (funcall form)))
       ;; Non-interactive invocation.
@@ -730,9 +744,7 @@ All args are optional, the current button and buffer file are the defaults."
 (defun hui:help-ebut-highlight ()
   "Highlight any explicit buttons in help buffer associated with current buffer."
   (if (fboundp 'hproperty:but-create)
-      (save-excursion
-	(set-buffer
-	 (get-buffer (hypb:help-buf-name)))
+      (with-current-buffer (get-buffer (hypb:help-buf-name))
 	(hproperty:but-create))))
 
 (defun hui:htype-delete (htype-sym)
@@ -754,18 +766,21 @@ Optional NO-SORT means display in decreasing priority order (natural order)."
 	      (tprefix (concat tstr "::"))
 	      (buf-name (hypb:help-buf-name (capitalize tstr)))
 	      (temp-buffer-show-hook
-	       (function
-		 (lambda (buf)
-		  (set-buffer buf) (goto-char (point-min))
-		  (replace-regexp "^" "  ") (goto-char (point-min))
-		  (replace-string (concat "  " tprefix) "") 
-		  (goto-char (point-min)) (set-buffer-modified-p nil)
-		  (display-buffer buf nil))))
+	       (lambda (buf)
+		 (set-buffer buf) (goto-char (point-min))
+		 (while (re-search-forward "^" nil t)
+		   (replace-match "  " t nil))
+		 (goto-char (point-min))
+		 (while (search-forward (concat "  " tprefix) nil t)
+		   (replace-match "" t nil))
+		 (goto-char (point-min))
+		 (set-buffer-modified-p nil)
+		 (display-buffer buf nil)))
 	      (temp-buffer-show-function temp-buffer-show-hook)
 	      (names (htype:names htype-sym))
 	      (term (hargs:read-match
 		     (concat (capitalize tstr)
-			     " to describe (RTN for all): ")
+			     " to describe (RET for all): ")
 		     (mapcar 'list (cons "" names))
 		     nil t nil htype-sym))
 	      nm-list
@@ -773,24 +788,30 @@ Optional NO-SORT means display in decreasing priority order (natural order)."
 	 (setq nm-list
 	       (if (string= term "")
 		   (let ((type-names
-			   (mapcar (function (lambda (nm) (concat tprefix nm)))
+			   (mapcar (lambda (nm) (concat tprefix nm))
 				   names)))
 		     (if no-sort type-names
 		       (sort type-names 'string<)))
 		 (cons (concat tprefix term) nil))
 	       doc-list (delq nil (mapcar
-				    (function
-				      (lambda (name)
-					(let ((doc (documentation
-						     (intern-soft name))))
-					  (if doc (cons name doc)))))
+				    (lambda (name)
+				      (let ((doc (documentation
+						  (intern-soft name))))
+					(if doc (cons name doc))))
 				    nm-list)))
 	 (with-output-to-temp-buffer buf-name
-	   (mapcar (function (lambda (nm-doc-cons)
-			       (princ (car nm-doc-cons)) (terpri)
-			       (princ (cdr nm-doc-cons)) (terpri)))
+	   (mapcar (lambda (nm-doc-cons)
+		     (princ (car nm-doc-cons)) (terpri)
+		     (princ (cdr nm-doc-cons)) (terpri))
 		   doc-list)))))
 
+(defun hui:htype-help-current-window (htype-sym &optional no-sort)
+  "Displays in the current window, documentation for types from HTYPE-SYM which match to a regexp.
+Optional NO-SORT means display in decreasing priority order (natural order)."
+  (let ((display-buffer-alist
+	 '(("\\`*Help" . ((lambda (buf _alist) (switch-to-buffer buf)))))))
+    (hui:htype-help htype-sym no-sort)))
+
 (defun hui:key-dir (but-buf)
   "Returns button key src directory based on BUT-BUF, a buffer."
   (if (bufferp but-buf)
@@ -804,19 +825,19 @@ Optional NO-SORT means display in decreasing priority order (natural order)."
   "Returns button key src location based on BUT-BUF, a buffer.
 This is BUT-BUF when button data is stored in the buffer and the
 button's source file name when the button data is stored externally."
-  (save-excursion
-    (set-buffer but-buf)
+  (with-current-buffer but-buf
     (cond ((hmail:mode-is-p) but-buf)
 	  ((hpath:symlink-referent (buffer-file-name but-buf)))
 	  (t but-buf))))
 
 (defun hui:link-create (modify but-window lbl-key but-loc but-dir type-and-args)
-  "Creates or modifies a new Hyperbole button.
+  "Creates or modifies a new Hyperbole link button.
 If MODIFY is non-nil, modifies button at point in BUT-WINDOW,
 otherwise, prompts for button label and creates a button.
 LBL-KEY is internal form of button label.  BUT-LOC is file or buffer
 in which to create button.  BUT-DIR is directory of BUT-LOC.
-TYPE-AND-ARGS is the action type for the button followed by any arguments it requires."
+TYPE-AND-ARGS is the action type for the button followed by any
+arguments it requires.  Any text properties are removed from string arguments." 
   (hattr:set 'hbut:current 'loc but-loc)
   (hattr:set 'hbut:current 'dir but-dir)
   (hattr:set 'hbut:current 'actype (intern-soft
@@ -824,54 +845,58 @@ TYPE-AND-ARGS is the action type for the button followed by any arguments it req
 					     (symbol-name
 					       (car type-and-args)))))
   (hattr:set 'hbut:current 'args (cdr type-and-args))
-
   (select-window but-window)
   (let ((label (ebut:key-to-label lbl-key)))
-    (ebut:operate label (if modify label)))
-  )
+    (ebut:operate label (if modify label))))
 
 (defun hui:link-possible-types ()
   "Returns list of possible link types for a Hyperbole button link to point.
 Each list element is a list of the link type and any arguments it requires.
 
-The link types considered are fixed.  Defining new link types will not alter
-the possible types.  The code must be changed to do that.
+The link types considered are fixed; this function must be changed to alter
+the contexts recognized.  Defining new link types will not alter the
+possible types.
 
 Referent Context         Possible Link Type Returned
 ----------------------------------------------------
 Explicit Button          link-to-ebut
-  or
+Info Index Item          link-to-Info-index-item
 Info Node                link-to-Info-node
-  or
-Mail Reader Msg          link-to-mail
-
-Outline Regexp Prefix    link-to-string-match
-  or
+Mail Reader Message      link-to-mail
 Directory Name           link-to-directory
-  or
 File Name                link-to-file
-  or
 Koutline Cell            link-to-kcell
-  or
+Outline Heading          link-to-string-match
 Buffer attached to File  link-to-file
-  or
-Buffer without File      link-to-buffer-tmp
+Buffer without File      link-to-buffer-tmp"
+;; Elisp Buffer at Start
+;; or End of Sexpression    eval-elisp
 
-Elisp Buffer at Start
-or End of Sexpression    eval-elisp
-"
   (let (val)
     (delq nil
 	  (list (if (ebut:at-p)
 		    (list 'link-to-ebut buffer-file-name (ebut:label-p)))
 		(cond ((eq major-mode 'Info-mode)
-		       (let ((hargs:reading-p 'Info-node))
-			 (list 'link-to-Info-node (hargs:at-p))))
+		       (if (and Info-current-node
+				(member Info-current-node
+				       (Info-index-nodes Info-current-file))
+				(Info-menu-item-at-p))
+			   (let ((hargs:reading-p 'Info-index-item))
+			     (list 'link-to-Info-index-item (hargs:at-p)))
+			 (let ((hargs:reading-p 'Info-node))
+			   (list 'link-to-Info-node (hargs:at-p)))))
 		      ((hmail:reader-p)
 		       (list 'link-to-mail
-			     (list (rmail:msg-id-get) buffer-file-name)))
-		      )
+			     (list (rmail:msg-id-get) buffer-file-name))))
 		(cond
+		 ((let ((hargs:reading-p 'directory))
+		    (setq val (hargs:at-p t)))
+		  (list 'link-to-directory val))
+		 ((let ((hargs:reading-p 'file))
+		    (setq val (hargs:at-p t)))
+		  (list 'link-to-file val (point)))
+		 ((eq major-mode 'kotl-mode)
+		  (list 'link-to-kcell buffer-file-name (kcell-view:idstamp)))
 		 ;; If link is within an outline-regexp prefix, use
 		 ;; a link-to-string-match.
 		 ((and (boundp 'outline-regexp)
@@ -885,7 +910,7 @@ or End of Sexpression    eval-elisp
 				 0)))))
 		  (save-excursion
 		    (end-of-line)
-		    (let ((heading (buffer-substring
+		    (let ((heading (buffer-substring-no-properties
 				    (point)
 				    (progn (beginning-of-line) (point))))
 			  (occur 1))
@@ -893,30 +918,25 @@ or End of Sexpression    eval-elisp
 			(setq occur (1+ occur)))
 		      (list 'link-to-string-match
 			    heading occur buffer-file-name))))
-		 ((let ((hargs:reading-p 'directory))
-		    (setq val (hargs:at-p t)))
-		  (list 'link-to-directory val))
-		 ((let ((hargs:reading-p 'file))
-		    (setq val (hargs:at-p t)))
-		  (list 'link-to-file val (point)))
-		 ((eq major-mode 'kotl-mode)
-		  (list 'link-to-kcell buffer-file-name (kcell-view:idstamp)))
 		 (buffer-file-name
 		  (list 'link-to-file buffer-file-name (point)))
-		 (t (list 'link-to-buffer-tmp (buffer-name)))
-		 )
-		(and (fboundp 'smart-emacs-lisp-mode-p)
-		     (smart-emacs-lisp-mode-p)
-		     (or (= (char-syntax (following-char)) ?\()
-			 (= (char-syntax (preceding-char)) ?\)))
-		     (setq val (hargs:sexpression-p))
-		     (list 'eval-elisp val))
+		 (t (list 'link-to-buffer-tmp (buffer-name))))
+		;; (and (fboundp 'smart-emacs-lisp-mode-p)
+		;;      (smart-emacs-lisp-mode-p)
+		;;      (or (eq (char-syntax (following-char)) ?\()
+		;; 	 (eq (char-syntax (preceding-char)) ?\)))
+		;;      (setq val (hargs:sexpression-p))
+		;;      (list 'eval-elisp val))
 		))))
 
+(defun hui:list-remove-text-properties (lst)
+  "Returns LST, a list, with text properties removed from any string elements."
+  (mapcar (lambda (elt) (if (stringp elt) (substring-no-properties elt) elt))
+	  lst))
 
-;;;
+;;; ************************************************************************
 ;;; Private variables
-;;;
+;;; ************************************************************************
 
 
 (defvar hui:ebut-label-prev nil
